$VAR1 = {
          'File::Spec' => 0,
          'Exporter::Tidy' => 0,
          'POSIX' => 0,
          'File::Glob' => 0,
          'Data::Dumper' => 0,
          'Env' => 0,
          'AutoLoader' => 0,
          'Cwd' => 0
        };
