package Bundle::Zoidberg;

our $VERSION = 0.04;

__END__

=head1 NAME

Bundle::Zoidberg - A bundle to install the zoidberg shell

=head1 SYNOPSIS

 perl -MCPAN -e 'install Bundle::Zoidberg'

=head1 DESCRIPTION

This bundle contains the zoidberg shell with all bells and whistles.
Dependencies are not listed.

=head1 CONTENTS

Env::PS1 0.04 - fancy prompt

Term::ReadLine::Zoid 0.05 - input module

Zoidberg - the shell itself

=head1 AUTHOR

Jaap Karssenberg E<lt>pardus@cpan.orgE<gt>

=cut
